/*
 *	FM-7 EMULATOR "XM7"
 *
 *	Copyright (C) 1999-2019 ohD(Twitter:@xm6_original)
 *	Copyright (C) 2001-2019 Ryu Takegami (Twitter:@RyuTakegami)
 *	Copyright (C) 2010-2018 Toma
 *
 *	[ ROM (񊿎EꐅE񐅏) ]
 */

#include <stdlib.h>
#include <string.h>
#include "xm7.h"
#include "kanji.h"
#include "subctrl.h"
#include "device.h"
#include "jcard.h"
#include "jsubsys.h"

/*
 *	O[o [N
 */
WORD kanji_addr;						/* AhXWX^() */
BYTE *kanji_rom;						/* PROM(JIS78) */
int kanji_rom_type;						/* PROM(JIS78) */
#if XM7_VER >= 2
BYTE *kanji_rom1;						/* PROM(JIS83) */
int kanji_rom1_type;					/* PROM(JIS83) */
#endif
BYTE *kanji_rom2;						/* QROM */
int kanji_rom2_type;					/* QROM */
#if XM7_VER >= 3 || defined(JSUB)
BYTE *kanji_romn;						/* PROM for {ʐMJ[h/FM77AV40EXROM */
int kanji_romn_type;					/* PROM for {ʐMJ[h/FM77AV40EXROM */
#endif
#if XM7_VER == 1
BOOL kanji_enable;						/* ROMLtO for FM-7/8 */
#endif
BOOL kanji_asis_flag;					/* JIS78G~[VtO */

/*
 *	vg^Cv錾
 */
static void FASTCALL kanji_make_jis78(BYTE *rom);

/*
 *	ID
 */
#define	ID	MAKEID('K', 'A', 'N', 'J')
//#define KANJI_LOG

/*
 *	ROM
 *	
 */
BOOL FASTCALL kanji_init(void)
{
	/*  */
#if XM7_VER == 1
	kanji_enable = TRUE;
#endif
	kanji_rom_type = KANJI_ROM_NONE;
#if XM7_VER >= 2
	kanji_rom1_type = KANJI_ROM_NONE;
	kanji_rom1 = NULL;
#endif
	kanji_rom2_type = KANJI_ROM_NONE;
	kanji_rom2 = NULL;
#if XM7_VER >= 3 || defined(JSUB)
	kanji_romn_type = KANJI_ROM_NONE;
	kanji_romn = NULL;
#endif

	/* m */
#if XM7_VER >= 2
	kanji_rom = (BYTE *)malloc(0x80000);
	kanji_rom1 = (BYTE *)(kanji_rom + 0x20000);
#elif defined(JSUB)
	kanji_rom = (BYTE *)malloc(0x60000);
#else
	kanji_rom = (BYTE *)malloc(0x40000);
#endif
	if (!kanji_rom) {
		return FALSE;
	}

#if XM7_VER >= 2
	/* JIS1 t@Cǂݍ (KANJIA.ROMKANJI.ROMKANJI1.ROM) */
	if (file_load(KANJI_ROMA, kanji_rom, 0x20000)) {
		kanji_rom_type = KANJI_ROM_JIS78L1A;
	}
	else if (file_load(KANJI_ROM, kanji_rom, 0x20000)) {
		kanji_rom_type = KANJI_ROM_JIS78L1;
	}
#else
	/* JIS1 t@Cǂݍ (KANJI.ROMKANJIA.ROMKANJI1.ROM) */
	if (file_load(KANJI_ROM, kanji_rom, 0x20000)) {
		kanji_rom_type = KANJI_ROM_JIS78L1;
	}
	else if (file_load(KANJI_ROMA, kanji_rom, 0x20000)) {
		kanji_rom_type = KANJI_ROM_JIS78L1A;
	}
#endif
	else if (file_load(KANJI_ROM1, kanji_rom, 0x20000)) {
		kanji_rom_type = KANJI_ROM_JIS83L1;
	}

#if XM7_VER >= 2
	/* JIS1(JIS83)tHg̓ǂݍ */
	if (kanji_rom_type != KANJI_ROM_JIS83L1) {
		if (file_load(KANJI_ROM1, kanji_rom1, 0x20000)) {
			/* JIS83 1ROM̃^Cvŕۑ */
			kanji_rom1_type = KANJI_ROM_JIS83L1;
		}
	}

	if (kanji_rom1_type == KANJI_ROM_NONE) {
#if XM7_VER >= 3
		if (kanji_rom_type != KANJI_ROM_JIS83L1) {
			/* V3łKANJI1.ROMȂꍇ̓G[ */
			return FALSE;
		}
#endif

		/* ɓǂݍ߂ROM̓e蓖Ă */
		memcpy(kanji_rom1, kanji_rom, 0x20000);

		/* 1ROM̖{̃^Cvŕۑ */
		kanji_rom1_type = kanji_rom_type;
	}
#endif

	/* JIS2 t@Cǂݍ */
#if XM7_VER >= 2
	kanji_rom2 = (BYTE *)(kanji_rom + 0x40000);
#else
	kanji_rom2 = (BYTE *)(kanji_rom + 0x20000);
#endif
	if (file_load(KANJI_ROM2, kanji_rom2, 0x20000)) {
		kanji_rom2_type = KANJI_ROM_JIS83L2;
	}
	else {
		/* 2ROMȂƂ */
		kanji_rom2_type = KANJI_ROM_NONE;
	}

#if XM7_VER >= 3 || defined(JSUB)
	/* JIS1 for {ʐMJ[h/FM77AV40EXgROM t@Cǂݍ */
#if XM7_VER >= 2
	kanji_romn = (BYTE *)(kanji_rom + 0x60000);
#else
	kanji_romn = (BYTE *)(kanji_rom + 0x40000);
#endif
	if (file_load(KANJI_ROMN, kanji_romn, 0x20000)) {
		kanji_romn_type = KANJI_ROM_JIS78L1N;

		/* KANJI.ROMnǂݍ߂ĂȂKANJIN.ROM̓eRs[ */
		if (kanji_rom_type == KANJI_ROM_NONE) {
			memcpy(kanji_rom, kanji_romn, 0x20000);
			kanji_rom_type = KANJI_ROM_JIS78L1N;
		}
	}
	else {
		/* łȂΊɓǂݍݍς݂JIS78tHg] */
		if (kanji_rom_type != KANJI_ROM_JIS83L1) {
			memcpy(kanji_romn, kanji_rom, 0x20000);
			kanji_romn_type = kanji_rom_type;
		}
	}
#endif

	/* ROMG~[V̉ۂJIS78tHg̗LŔf */
#ifdef JSUB
	if (kanji_rom_type != KANJI_ROM_NONE) {
		if ((kanji_romn_type == KANJI_ROM_NONE) ||
			(kanji_rom_type == KANJI_ROM_JIS83L1)) {
			/* {ʐMJ[h̊ROMȂꍇ */
			if (kanji_romn_type == KANJI_ROM_NONE) {
				memcpy(kanji_romn, kanji_rom, 0x20000);

				/* JIS78ROMf[^ */
				kanji_make_jis78(kanji_romn);
				kanji_romn_type = KANJI_ROM_JIS78L1E;
				if (kanji_rom_type != KANJI_ROM_JIS83L1) {
					kanji_rom_type = KANJI_ROM_JIS78L1E;
					memcpy(kanji_rom, kanji_romn, 0x20000);
				}
			}
			else {
				/* JIS78ROMf[^ */
				kanji_make_jis78(kanji_rom);
				kanji_rom_type = KANJI_ROM_JIS78L1E;
			}
		}
	}
#else
	/* JIS78ROMf[^ */
	if (kanji_rom_type == KANJI_ROM_JIS83L1) {
		kanji_make_jis78(kanji_rom);
		kanji_rom_type = KANJI_ROM_JIS78L1E;
	}
#endif

#if XM7_VER >= 2
#if XM7_VER >= 3
	if ((kanji_rom_type == KANJI_ROM_NONE) ||
		(kanji_rom2_type != KANJI_ROM_JIS83L2)) {
#else
	if (kanji_rom_type == KANJI_ROM_NONE) {
#endif
		return FALSE;
	}
#endif

	return TRUE;
}

/*
 *	ROM
 *	N[Abv
 */
void FASTCALL kanji_cleanup(void)
{
	if (kanji_rom) {
		free(kanji_rom);
		kanji_rom = NULL;
#if XM7_VER >= 2
		kanji_rom1 = NULL;
#endif
		kanji_rom2 = NULL;
#if XM7_VER >= 3 || defined(JSUB)
		kanji_romn = NULL;
#endif
	}
}

/*
 *	ROM
 *	Zbg
 */
void FASTCALL kanji_reset(void)
{
	LOG0(Normal, "Zbg");

	kanji_addr = 0;
}

/*
 *	ROM
 *	PoCgǂݏo
 */
BOOL FASTCALL kanji_readb(WORD addr, BYTE *dat)
{
	int offset;

	switch (addr) {
		/* AhX(1E2)ʁE */
		case 0xfd20:
		case 0xfd21:
#if XM7_VER >= 2
		case 0xfd2c:
		case 0xfd2d:
#endif
			*dat = 0xff;
			return TRUE;

		/* 1f[^ */
		case 0xfd22:		/* 1f[^LEFT */
		case 0xfd23:		/* 1f[^RIGHT */
#if XM7_VER == 1
			/* ROMJ[hȂ */
			if ((fm_model == MODEL_FM7) && !kanji_enable) {
#ifdef JSUB
				/* ROM̊ROM$FD20ɔzuĂȂ */
				if (!(jsub_enable && jsub_kanji_dipsw)) {
					*dat = 0xff;
					return TRUE;
				}
#else
				*dat = 0xff;
				return TRUE;
#endif
			}

			/* LN^ZbgȂ */
			if ((fm_model == MODEL_FM8) && !kanji_enable) {
				*dat = 0xff;
				return TRUE;
			}

			/* ROMgp\ȏԂł͂Ȃ */
			if (!((fm_model == MODEL_FM7) &&
				 jsub_enable && jsub_kanji_dipsw) &&
				 (kanji_rom_type == KANJI_ROM_NONE)) {
				*dat = 0xff;
				return TRUE;
			}
#endif

#if XM7_VER >= 3 || (XM7_VER == 1 && defined(L4CARD))
			/* ROMTuVXeɐڑ */
			if (subkanji_flag) {
				*dat = 0xff;
				return TRUE;
			}
#endif

			offset = kanji_addr << 1;
#ifdef JSUB
			if ((fm_model < MODEL_FM77AV) &&
				(offset >= 0x6000) && (offset < 0x8000) &&
				!kanji_asis_flag) {
				/* FM-7[h$6000`$7FFF͖`̈ */
				*dat = (BYTE)(addr & 1);
			}
			else {
				if ((fm_model <= MODEL_FM77AV) ||
					((fm_model == MODEL_FM7) && 
					jsub_enable && jsub_kanji_dipsw)) {
					*dat = kanji_romn[offset + (addr & 1)];
				}
				else {
#if XM7_VER >= 2
					if ((fm_model >= MODEL_77AV20) &&
						(kanji_rom1_type == KANJI_ROM_JIS83L1)) {
						*dat = kanji_rom1[offset + (addr & 1)];
					}
					else {
						*dat = kanji_rom[offset + (addr & 1)];
					}
#else
					*dat = kanji_rom[offset + (addr & 1)];
#endif
				}
			}
#else
			*dat = kanji_rom[offset + (addr & 1)];
#endif
			return TRUE;

#if XM7_VER >= 2
		/* 2f[^ */
		case 0xfd2e:		/* 2f[^LEFT */
		case 0xfd2f:		/* 2f[^RIGHT */
#if XM7_VER >= 3
			if ((fm_model < MODEL_FM77AV) ||
				((fm_model <= MODEL_AV20EX) && (jcard_mode != JCARD_ON))) {
#else
			if ((fm_model < MODEL_FM77AV) || (jcard_mode != JCARD_ON)) {
#endif
				return FALSE;
			}

#if XM7_VER >= 3
			/* ROMTuVXeɐڑ */
			if (subkanji_flag) {
				*dat = 0xff;
				return TRUE;
			}
#endif

			offset = kanji_addr << 1;
			*dat = kanji_rom2[offset + (addr & 1)];
			return TRUE;
#endif
	}

	return FALSE;
}

/*
 *	ROM
 *	PoCg
 */
BOOL FASTCALL kanji_writeb(WORD addr, BYTE dat)
{
	switch (addr) {
		/* AhX(1E2) */
#if XM7_VER >= 2
		case 0xfd2c:
#if XM7_VER >= 3
			if ((fm_model < MODEL_FM77AV) ||
				((fm_model <= MODEL_AV20EX) && (jcard_mode != JCARD_ON))) {
#else
			if ((fm_model < MODEL_FM77AV) || (jcard_mode != JCARD_ON)) {
#endif
				return FALSE;
			}
#endif
		case 0xfd20:
			kanji_addr &= 0x00ff;
			kanji_addr |= (WORD)(dat << 8);
			return TRUE;

		/* AhX(1E2) */
#if XM7_VER >= 3 || (XM7_VER == 2 && defined(JCARD))
		case 0xfd2d:
#if XM7_VER >= 3
			if ((fm_model < MODEL_FM77AV) ||
				((fm_model <= MODEL_AV20EX) && (jcard_mode != JCARD_ON))) {
#else
			if ((fm_model < MODEL_FM77AV) || (jcard_mode != JCARD_ON)) {
#endif
				return FALSE;
			}
#endif
		case 0xfd21:
			kanji_addr &= 0xff00;
			kanji_addr |= dat;
			return TRUE;

		/* f[^ */
		case 0xfd22:
		case 0xfd23:
#if XM7_VER >= 2
		case 0xfd2f:
#endif
			return TRUE;

#if XM7_VER >= 2
		/* {J[h oNZNg */
		case 0xfd2e:
#if XM7_VER >= 3
			if ((fm_model < MODEL_FM77AV) ||
				((fm_model <= MODEL_AV20EX) && (jcard_mode != JCARD_ON))) {
#else
			if ((fm_model < MODEL_FM77AV) || (jcard_mode != JCARD_ON)) {
#endif
				return FALSE;
			}

			/* bit7:wKRAM */
			if (dat & 0x80) {
				dicram_en = TRUE;
			}
			else {
				dicram_en = FALSE;
			}

			/* bit6:ROMCl[u */
			if (dat & 0x40) {
				dicrom_en = TRUE;
			}
			else {
				dicrom_en = FALSE;
			}

			/* bit0-5:ROMoN */
			dicrom_bank = (BYTE)(dat & 0x3f);
			return TRUE;
#endif
	}

	return FALSE;
}

/*
 *	ROM
 *	JIS78̈쐬
 */
static void FASTCALL kanji_make_jis78(BYTE *rom)
{
	/* JIS78`e[u(FM77AV) */
	static const DWORD jis78_table[] = {
		0xffffffff, 0x00000001, 0xffff8001, 0xfc00ffff,
		0x00000001, 0x00000001, 0xfe000001, 0x00000001,
		0xffffffff, 0x80000000, 0xffffffff, 0xf8000001,
		0xfff00000, 0xff800000, 0xffffffff, 0xfffc0000,
		0xffffffff, 0x00000000, 0xffffffff, 0xf8000001,
		0x00000000, 0x00000000, 0xfe000001, 0x0001fffc,
	};

	/* JIS781JIS832֊e[u */
	static const DWORD convert_table[][2] = {
		/* 1code addr		  2code addr */
		{/* 0x3033*/0x4130,	/* 0x724D*/0xE4D0},
		{/* 0x3229*/0x4490,	/* 0x7274*/0xD540},
		{/*a 0x3342*/0x6620,	/*y 0x695A*/0x93A0},
		{/*h 0x3349*/0x6690,	/* 0x5978*/0x5380},
		{/* 0x3376*/0x8760,	/*} 0x635E*/0x87E0},
		{/* 0x3443*/0x6830,	/* 0x5E75*/0x5D50},
		{/* 0x3452*/0x6920,	/*| 0x6B5D*/0x97D0},
		{/*z 0x375B*/0x6FB0,	/* 0x7074*/0xD140},
		{/*{ 0x395C*/0x73C0,	/* 0x6268*/0xA480},
		{/* 0x3C49*/0x7890,	/*A 0x6922*/0x7220},
		{/*x 0x3F59*/0x7F90,	/* 0x7057*/0xE170},
		{/*G 0x4128*/0xA280,	/* 0x6C4D*/0x98D0},
		{/* 0x445B*/0xC9B0,	/* 0x5464*/0x4840},
		{/*v 0x4557*/0xCB70,	/* 0x626A*/0xA4A0},
		{/* 0x456E*/0xEAE0,	/* 0x5B6D*/0x56D0},
		{/* 0x4573*/0xEB30,	/* 0x5E39*/0x1D90},
		{/* 0x4676*/0xED60,	/* 0x6D6E*/0xBAE0},
		{/* 0x4768*/0xEE80,	/* 0x6A24*/0x7440},
		{/*O 0x4930*/0xB300,	/*w 0x5B58*/0x3780},
		{/* 0x4B79*/0xF790,	/* 0x5056*/0x2160},
		{/* 0x4C79*/0xF990,	/*M 0x692E*/0x72E0},
		{/*U 0x4F36*/0xBF60,	/* 0x6446*/0x8860},
		{/* 0x3646*/0x6C60,	/* 0x7421*/0xC810},
		{/* 0x4B6A*/0xF6A0,	/* 0x7422*/0xC820},
		{/*y 0x4D5A*/0xDBA0,	/* 0x7423*/0xC830},
		{0, 0},
	};

#if XM7_VER >= 2
	/* JIS78 for FM77AVJIS83 񊿎փe[u */
	static const DWORD convert_table_av[][3] = {
	/*	  JIS78    JIS83   words	*/
		{0x06020, 0x06c20, 26},
		{0x06800, 0x00d40, 6},
		{0x068c0, 0x00f40, 5},
		{0x07020, 0x041c0, 4},
		{0x070a0, 0x04260, 1},
		{0x070c0, 0x042a0, 1},
		{0x070e0, 0x04240, 1},
		{0x07100, 0x04280, 1},
		{0x07120, 0x042c0, 1},
		{0x07140, 0x04180, 2},
//		{0x07820, 0, 0},
//		{0x078a0, 0, 0},
//		{0x07980, 0, 0},
		{0, 0, 0},
	};
#endif

	DWORD addr;
	DWORD i;
#if XM7_VER >= 2
	DWORD j;
#endif

	/* pb`ĕsvtOĂꍇ͏XLbv */
	if (kanji_asis_flag) {
		return;
	}

#if XM7_VER >= 2
	/* JIS78񊿎(AVƎ)JIS83񊿎璊o */
	for (addr = 0; convert_table_av[addr][0]; addr ++) {
		for (j = 0; j < convert_table_av[addr][2]; j ++) {
			/* tHgRs[ */
			for (i = 0; i < 32; i ++) {
				rom[convert_table_av[addr][0] + i] =
				rom[convert_table_av[addr][1] + i];
			}
		}
	}
#endif

	/* JIS78ROMł͖`ɂȂĂ镔𖢒`Ƃ */
	for (addr = 0; addr < 0x6000; addr += 32) {
		if (jis78_table[addr >> 10] & (1 << (addr >> 5))) {
			/* `tHg쐬 */
			for (i = 0; i < 32; i += 2) {
				rom[addr + i + 0] = 0x00;
				rom[addr + i + 1] = 0x01;
			}
		}
	}

#if XM7_VER >= 2 || defined(JSUB)
	/* KANJI2.ROMȂꍇ͂̂܂܋A */
	if (kanji_rom2_type != KANJI_ROM_JIS83L2) {
		return;
	}

	/* JIS83œւꂽ2ROM㏑ */
	for (addr = 0; convert_table[addr][0]; addr ++) {
		/* tHgRs[ */
		for (i = 0; i < 32; i ++) {
			rom[convert_table[addr][0] * 2 + i] =
				kanji_rom2[convert_table[addr][1] * 2 + i];
		}
	}
#endif
}

/*
 *	ROM
 *	Z[u
 */
BOOL FASTCALL kanji_save(int fileh)
{
	LOG0(Normal, "Z[u");

	if (!file_word_write(fileh, kanji_addr)) {
		return FALSE;
	}
#if XM7_VER == 1
	if (!file_bool_write(fileh, kanji_enable)) {
		return FALSE;
	}
#else
	if (!file_bool_write(fileh, TRUE)) {
		return FALSE;
	}
#endif

	return TRUE;
}

/*
 *	ROM
 *	[h
 */
BOOL FASTCALL kanji_load(int fileh, int ver)
{
#if XM7_VER >= 2
	BOOL tmp;
#endif

	/* o[W`FbN */
	if (ver < 200) {
		return FALSE;
	}

	LOG0(Normal, "[h");

	if (!file_word_read(fileh, &kanji_addr)) {
		return FALSE;
	}
#if XM7_VER >= 3
	if ((ver < 1201) || ((ver >= 1400) && (ver < 1401))) {
#elif XM7_VER >= 2
	if (ver < 1201) {
#else
	if (ver < 1001) {
#endif
		return TRUE;
	}
#if XM7_VER == 1
	if (!file_bool_read(fileh, &kanji_enable)) {
#else
	if (!file_bool_read(fileh, &tmp)) {
#endif
		return FALSE;
	}

	return TRUE;
}

