/*
 *	FM-7 EMULATOR "XM7"
 *
 *	Copyright (C) 1999-2019 ohD(Twitter:@xm6_original)
 *	Copyright (C) 2001-2019 Ryu Takegami (Twitter:@RyuTakegami)
 *	Copyright (C) 2010-2018 Toma
 *
 *	[ XPW[ ]
 */

#include <string.h>
#include "xm7.h"
#include "jsubsys.h"
#include "subctrl.h"
#include "display.h"
#include "mmr.h"
#include "fdc.h"
#include "jsubsys.h"
#if XM7_VER >= 2 || (XM7_VER == 1 && defined(SFDC) && defined(IFDC))
#include "dmac.h"
#endif
#include "device.h"
#include "event.h"
#include "tapelp.h"

/*
 *	O[o [N
 */
BOOL run_flag;							/* 쒆tO */
BOOL stopreq_flag;						/* ~vtO */
event_t event[EVENT_MAXNUM];			/* Cxg f[^ */
breakp_t breakp[BREAKP_MAXNUM];			/* u[N|Cg f[^ */
WORD main_runadr;						/* CCPȖOsAhX */
WORD sub_runadr;						/* TuCPȖOsAhX */
DWORD main_speed;						/* CCPUXs[h */
DWORD mmr_speed;						/* C(MMR)Xs[h */
#if XM7_VER >= 2
DWORD fmmr_speed;						/* C(MMR)Xs[h */
#endif
DWORD sub_speed;						/* TuCPUXs[h */
WORD main_overcycles;					/* CCPUI[o[TCN */
WORD sub_overcycles;					/* TuCPUI[o[TCN */
BOOL cycle_steal;						/* TCNX`[tO */
#if XM7_VER == 1
BOOL cycle_steal_mode;					/* TCNX`[XCb` */
BOOL cycle_steal_lock;					/* TCNX`[ύXs */
#endif
BOOL subclock_mode;						/* TuCPU uLO^C~O */
DWORD speed_ratio;						/* CPU쑬x(%) */
DWORD vmtime;							/* VMz */
#if XM7_VER == 1
DWORD main_speed_low;					/* CCPUXs[h(ᑬ) */
DWORD sub_speed_low;					/* TuCPUXs[h(ᑬ) */
#endif
#ifdef JSUB
DWORD jsub_speed;						/* {TuCPUXs[h */
WORD jsub_overcycles;					/* {TuCPUI[o[TCN */
WORD jsub_runadr;						/* {TuCPȖOsAhX */
#endif
#if XM7_VER == 1 && defined(Z80CARD)
WORD mainz80_runadr;					/* CCPU(Z80)̑OsAhX */
#endif

/*
 *	X^eBbN [N
 */
static BOOL break_flag;					/* u[N|CgLtO */

/*
 *	ID
 */
#define	ID	MAKEID('S', 'C', 'H', 'E')
//#define SCHEDULE_LOG

/*
 *	XPW[
 *	
 */
BOOL FASTCALL schedule_init(void)
{
	run_flag = FALSE;
	stopreq_flag = FALSE;
	break_flag = FALSE;
	memset(breakp, 0, sizeof(breakp));
	memset(event, 0, sizeof(event));

	/* O̎sAhX */
	main_runadr = 0xFFFF;
	sub_runadr = 0xFFFF;
#ifdef JSUB
	jsub_runadr = 0xFFFF;
#endif
#if XM7_VER == 1 && defined(Z80CARD)
	mainz80_runadr = 0xFFFF;
#endif

	/* CPUxݒ */
	main_speed = MAINCYCLES * 10;
	mmr_speed = MAINCYCLES_MMR * 10;
#if XM7_VER >= 2
	fmmr_speed = MAINCYCLES_FMMR * 10;
#endif
	sub_speed = SUBCYCLES * 10;
	cycle_steal = TRUE;
#if XM7_VER == 1
	cycle_steal_mode = TRUE;
	cycle_steal_lock = FALSE;
#endif
#if XM7_VER == 1
	subclock_mode = TRUE;
#else
	subclock_mode = FALSE;
#endif
	main_overcycles = 0;
	sub_overcycles = 0;
#if XM7_VER == 1
	main_speed_low = MAINCYCLES_LOW * 10;
	sub_speed_low = SUBCYCLES_LOW * 10;
#endif
#ifdef JSUB
	jsub_speed = JSUBCYCLES * 10;
	jsub_overcycles = 0;
#endif

	/* zԏ */
	vmtime = 0;

	return TRUE;
}

/*
 *	XPW[
 *	N[Abv
 */
void FASTCALL schedule_cleanup(void)
{
}

/*
 *	XPW[
 *	Zbg
 */
void FASTCALL schedule_reset(void)
{
	int i;

	LOG0(Normal, "Zbg");

	/* RTCCxgȊÕXPW[NA */
	/* (XM7NRTCCxg schedule_init ŃNAĂ) */
	for (i=0; i<EVENT_MAXNUM; i++) {
#if XM7_VER >= 2
		if (i != EVENT_RTC) {
			memset(&event[i], 0, sizeof(event_t));
		}
#else
		memset(&event[i], 0, sizeof(event_t));
#endif
	}

	/* JE^NA */
	maincpu.total = 0;
	subcpu.total = 0;
	main_overcycles = 0;
	sub_overcycles = 0;
#ifdef JSUB
	jsubcpu.total = 0;
	jsub_overcycles = 0;
#endif

	/* zԏ */
	vmtime = 0;

	/* sx䗦 */
	speed_ratio = 10000;
}

/*
 *	XPW[
 *	sTCN擾
 */
DWORD FASTCALL schedule_get_cycle(void)
{
	DWORD tmp;

#if XM7_VER >= 2
	if (mmr_fastmode) {
		tmp = fmmr_speed;
	}
	else
#endif
#if XM7_VER == 1
	if (lowspeed_mode ||
		((fm_model == MODEL_FM8) && tape_motor && tape_lowspeed)) {
		tmp = main_speed_low;
	}
	else {
		if ((mmr_flag || twr_flag) && (mmr_extrmw != EXRAM_960KB)) {
#else
	{
		if (mmr_flag || twr_flag) {
#endif
			tmp = mmr_speed;
		}
		else {
			tmp = main_speed;
		}
	}

#if XM7_VER >= 2
	/* tbV[h̃TCN𐶐 */
	if (!mmr_fastmode && mmr_fast_refresh) {
		if (mmr_flag || twr_flag) {
			/* MMR/TWRL 8.9%Abv */
			tmp = (DWORD)((tmp * 4461) >> 12);
		}
		else {
			/* MMR/TWR 8.6%Abv */
			tmp = (DWORD)((tmp * 4447) >> 12);
		}
	}
#endif

	/* CPUx䗦 */
	if (speed_ratio != 10000) {
		tmp = (tmp * speed_ratio) / 10000;
		if (tmp < 1) {
			tmp = 1;
		}
	}

	return tmp;
}

/*-[ Cxg ]-------------------------------------------------------------*/

/*
 *	XPW[
 *	Cxgݒ
 */
BOOL FASTCALL schedule_setevent(int id, DWORD microsec, BOOL (FASTCALL *func)(void))
{
	DWORD exec;

	ASSERT((id >= 0) && (id < EVENT_MAXNUM));
	ASSERT(func);

	if ((id < 0) || (id >= EVENT_MAXNUM)) {
		return FALSE;
	}
	if (microsec == 0) {
		event[id].flag = EVENT_NOTUSE;
		return FALSE;
	}

	/* o^ */
	event[id].current = microsec;
	event[id].reload = microsec;
	event[id].callback = func;
	event[id].flag = EVENT_ENABLED;

	/* sȂAԂ𑫂ĂKv(ň) */
	if (run_flag) {
		exec = (DWORD)maincpu.total;
		exec *= 10000;
		exec /= schedule_get_cycle();
		event[id].current += exec;
	}

	return TRUE;
}

/*
 *	XPW[
 *	Cxg폜
 */
BOOL FASTCALL schedule_delevent(int id)
{
	ASSERT((id >= 0) && (id < EVENT_MAXNUM));

	if ((id < 0) || (id >= EVENT_MAXNUM)) {
		return FALSE;
	}

	/* gp */
	event[id].flag = EVENT_NOTUSE;

	return TRUE;
}

/*
 *	XPW[
 *	Cxgnhݒ
 */
void FASTCALL schedule_handle(int id, BOOL (FASTCALL *func)(void))
{
	ASSERT((id >= 0) && (id < EVENT_MAXNUM));
	ASSERT(func);

	/* R[obN֐o^̂݁BȊO͐GȂ */
	event[id].callback = func;
}

/*
 *	XPW[
 *	ŒZsԒ
 */
static DWORD FASTCALL schedule_chkevent(DWORD microsec)
{
	DWORD exectime;
	int i;

	/* ݒ */
	exectime = microsec;

	/* CxgĒ */
	for (i=0; i<EVENT_MAXNUM; i++) {
		if (event[i].flag == EVENT_NOTUSE) {
			continue;
		}

		ASSERT(event[i].current > 0);
		ASSERT(event[i].reload > 0);

		if (event[i].current < exectime) {
			exectime = event[i].current;
		}
	}

	return exectime;
}

/*
 *	XPW[
 *	is
 */
static void FASTCALL schedule_doevent(DWORD microsec)
{
	int i;

	for (i=0; i<EVENT_MAXNUM; i++) {
		if (event[i].flag == EVENT_NOTUSE) {
			continue;
		}

		ASSERT(event[i].current > 0);
		ASSERT(event[i].reload > 0);

		/* sԂ */
		if (event[i].current < microsec) {
			event[i].current = 0;
		}
		else {
			event[i].current -= microsec;
		}

		/* JE^0Ȃ */
		if (event[i].current == 0) {
			/* ԂENABLE,DISABLEɂ炸[h */
			event[i].current = event[i].reload;
			/* R[obNs */
			if (event[i].flag == EVENT_ENABLED) {
				if (!event[i].callback()) {
					event[i].flag = EVENT_DISABLED;
				}
			}
		}
	}
}

/*-[ u[N|Cg ]-----------------------------------------------------*/

/*
 *	XPW[
 *	u[N|CgZbg(łɃZbgĂΏ)
 */
BOOL FASTCALL schedule_setbreak(int cpu, WORD addr)
{
	int i;

	/* ܂ASẴu[N|CgA邩 */
	for (i=0; i<BREAKP_MAXNUM; i++) {
		if (breakp[i].flag != BREAKP_NOTUSE) {
			if ((breakp[i].cpu == cpu) && (breakp[i].addr == addr)) {
				break;
			}
		}
	}
	/* ΁A폜 */
	if (i != BREAKP_MAXNUM) {
		breakp[i].flag = BREAKP_NOTUSE;
		/* u[NLtO`FbN */
		break_flag = FALSE;
		for (i=0; i<BREAKP_MAXNUM; i++) {
			if (breakp[i].flag != BREAKP_NOTUSE) {
				break_flag = TRUE;
			}
		}
		return TRUE;
	}

	/* 󂫂𒲍 */
	for (i=0; i<BREAKP_MAXNUM; i++) {
		if (breakp[i].flag == BREAKP_NOTUSE) {
			break;
		}
	}
	/* ׂĖ܂Ă邩 */
	if (i == BREAKP_MAXNUM) {
		return FALSE;
	}

	/* Zbg */
	breakp[i].flag = BREAKP_ENABLED;
	breakp[i].cpu = cpu;
	breakp[i].addr = addr;

	/* u[NL */
	break_flag = TRUE;

	return TRUE;
}

/*
 *	XPW[
 *	u[N|CgZbg(ʒuw)
 */
BOOL FASTCALL schedule_setbreak2(int num, int cpu, WORD addr)
{
	/* Zbg */
	if (breakp[num].flag != BREAKP_DISABLED) {
		breakp[num].flag = BREAKP_ENABLED;
	}
	breakp[num].cpu = cpu;
	breakp[num].addr = addr;

	/* u[NL */
	break_flag = TRUE;

	return TRUE;
}

/*
 *	XPW[
 *	u[N`FbN
 */
static BOOL FASTCALL schedule_chkbreak(void)
{
	int i;
	WORD main_prevrunadr;
	WORD sub_prevrunadr;
#ifdef JSUB
	WORD jsub_prevrunadr;
#endif
#if XM7_VER == 1 && defined(Z80CARD)
	WORD mainz80_prevrunadr;
#endif

	ASSERT(break_flag);

	/* A₱c */
#if XM7_VER == 1 && defined(Z80CARD)
	if (!main_z80mode) {
		main_prevrunadr = main_runadr;
		main_runadr = maincpu.pc;
	}
	else {
		mainz80_prevrunadr = mainz80_runadr;
		mainz80_runadr = (WORD)mainz80.pc;
	}
#else
	main_prevrunadr = main_runadr;
	main_runadr = maincpu.pc;
#endif
	sub_prevrunadr = sub_runadr;
	sub_runadr = subcpu.pc;
#ifdef JSUB
	jsub_prevrunadr = jsub_runadr;
	jsub_runadr = jsubcpu.pc;
#endif

	for (i=0; i<BREAKP_MAXNUM; i++) {
		if (breakp[i].flag == BREAKP_ENABLED) {
#if XM7_VER == 1 && defined(Z80CARD)
			/* CCPUZ80̏ꍇA */
			if ((breakp[i].cpu == MAINCPU) && !main_z80mode) {
#else
			if (breakp[i].cpu == MAINCPU) {
#endif
				if (breakp[i].addr == maincpu.pc) {
					if (maincpu.pc != main_prevrunadr) {
						return TRUE;
					}
				}
			}
			else if (breakp[i].cpu == SUBCPU) {
				if (breakp[i].addr == subcpu.pc) {
					if (subcpu.pc != sub_prevrunadr) {
						return TRUE;
					}
				}
			}
#ifdef JSUB
			else if (breakp[i].cpu == JSUBCPU) {
				if (breakp[i].addr == jsubcpu.pc) {
					if (jsubcpu.pc != jsub_prevrunadr) {
						return TRUE;
					}
				}
			}
#endif
#if XM7_VER == 1 && defined(Z80CARD)
			/* CCPUZ80̏ꍇ̂ݗL */
			else if ((breakp[i].cpu == MAINZ80) && main_z80mode) {
				if (breakp[i].addr == (WORD)mainz80.pc) {
					if ((WORD)mainz80.pc != mainz80_prevrunadr) {
						return TRUE;
					}
				}
			}
#endif
		}
	}

	return FALSE;
}

/*-[ s ]---------------------------------------------------------------*/

/*
 *	g[X
 */
void FASTCALL schedule_trace(void)
{
	/* Pߎs */
#if XM7_VER >= 2 || (XM7_VER == 1 && defined(SFDC) && defined(IFDC))
	if (!dma_burst_transfer || !dma_flag) {
		maincpu_execline();
	}
#else
	maincpu_execline();
#endif

#if XM7_VER >= 2 || (XM7_VER == 1 && defined(SFDC) && defined(IFDC))
	/* DMA] */
	if (dma_flag) {
		dmac_exec();
	}
#endif

	if ((!subhalt_flag || (subcpu.intr & INTR_HCF)) &&
		(cycle_steal || (subclock_mode || !(vrama_flag && !blank_flag)))) {
		subcpu_execline();
		/* VRAMANZXtOON̏ꍇ,vNbN1.83{ɂ */
		if (!cycle_steal && subclock_mode && vrama_flag) {
//			subcpu.total += (WORD)(subcpu.cycle * 2);
			subcpu.total += (WORD)(subcpu.cycle * 1.86f);
		}
	}
#ifdef JSUB
#if XM7_VER >= 2
	if (jsub_available && jsub_enable && !jsub_haltflag &&
		(fm_model <= MODEL_FM77AV)) {
#else
	if (jsub_available && jsub_enable && !jsub_haltflag &&
		(fm_model != MODEL_FM8)) {
#endif
		jsubcpu_execline();
	}
#endif

	/* HALTvɉ (V3.1) */
	subctrl_halt_ack();

	/* I[o[TCNNA */
	main_overcycles = 0;
	sub_overcycles = 0;
#ifdef JSUB
	jsub_overcycles = 0;
#endif
}

/*
 *	CS͋쓮
 */
void FASTCALL schedule_main_fullspeed(void)
{
	ASSERT(run_flag);
	ASSERT(!stopreq_flag);

#if XM7_VER >= 2 || (XM7_VER == 1 && defined(SFDC) && defined(IFDC))
	if (dma_burst_transfer) {
		/* o[Xg]sBCCPU͓Ȃ */
		return;
	}
#endif

	if (break_flag) {
		/* u[N|Cg */
		if (schedule_chkbreak()) {
			stopreq_flag = TRUE;
		}
		if (stopreq_flag) {
			return;
		}
	}

	/* CCPUs */
#ifdef FDDSND
#if XM7_VER == 1 && defined(SFDC)
	if (!fdc_waitmode || !(fdc_status[FDC_M] & FDC_ST_BUSY) || !(fdc_status[FDC_S] & FDC_ST_BUSY)) {
#else	/* XM7_VER == 1 && defined(SFDC) */
	if (!fdc_waitmode || !(fdc_status & FDC_ST_BUSY)) {
#endif	/* XM7_VER == 1 && defined(SFDC) */
		maincpu_exec();
	}
#else
	maincpu_exec();
#endif

	/* HALTvɉ (V3.1) */
	subctrl_halt_ack();

	/* I[o[TCNNA */
	main_overcycles = 0;
}

/*
 *	S͋쓮
 */
void FASTCALL schedule_fullspeed(void)
{
	ASSERT(run_flag);
	ASSERT(!stopreq_flag);

	if (break_flag) {
		/* u[N|Cg */
		if (schedule_chkbreak()) {
			stopreq_flag = TRUE;
		}
		if (stopreq_flag) {
			return;
		}
	}

	/* CCPUs */
#if XM7_VER >= 2 || (XM7_VER == 1 && defined(SFDC) && defined(IFDC))
	if (!dma_burst_transfer) {
#ifdef FDDSND
#if XM7_VER == 1 && defined(SFDC)
		if (!fdc_waitmode || !(fdc_status[FDC_M] & FDC_ST_BUSY) || !(fdc_status[FDC_S] & FDC_ST_BUSY)) {
#else	/* XM7_VER == 1 && defined(SFDC) */
		if (!fdc_waitmode || !(fdc_status & FDC_ST_BUSY)) {
#endif	/* XM7_VER == 1 && defined(SFDC) */
			maincpu_exec();
		}
#else
		maincpu_exec();
#endif
	}
#else
#ifdef FDDSND
#if XM7_VER == 1 && defined(SFDC)
	if (!fdc_waitmode || !(fdc_status[FDC_M] & FDC_ST_BUSY) || !(fdc_status[FDC_S] & FDC_ST_BUSY)) {
#else	/* XM7_VER == 1 && defined(SFDC) */
	if (!fdc_waitmode || !(fdc_status & FDC_ST_BUSY)) {
#endif	/* XM7_VER == 1 && defined(SFDC) */
		maincpu_exec();
	}
#else
	maincpu_exec();
#endif
#endif

	/* TuCPUs */
	if ((!subhalt_flag || (subcpu.intr & INTR_HCF)) &&
		(cycle_steal || (subclock_mode || !(vrama_flag && !blank_flag)))) {
		subcpu_execline();
		/* VRAMANZXtOON̏ꍇ,vNbN1.83{ɂ */
		if (!cycle_steal && subclock_mode && vrama_flag) {
			subcpu.total += (WORD)(subcpu.cycle * 1.86f);
		}
	}

#ifdef JSUB
	/* {TuCPUs */
#if XM7_VER >= 2
	if (jsub_available && jsub_enable && !jsub_haltflag &&
		(fm_model <= MODEL_FM77AV)) {
#else
	if (jsub_available && jsub_enable && !jsub_haltflag &&
		(fm_model != MODEL_FM8)) {
#endif
		jsubcpu_execline();
	}
#endif

	/* HALTvɉ (V3.1) */
	subctrl_halt_ack();

	/* I[o[TCNNA */
	main_overcycles = 0;
	sub_overcycles = 0;
#ifdef JSUB
	jsub_overcycles = 0;
#endif
}

/*
 *	s
 */
DWORD FASTCALL schedule_exec(DWORD microsec)
{
	DWORD exec;
	DWORD exec2;
	DWORD count;
	DWORD cycle;
	DWORD ratio;
	WORD main;
	WORD sub;
	DWORD limit;
	DWORD tmp;
#ifdef JSUB
	DWORD ratio_js;
	WORD jsub;
#endif

	/* ASSERT(run_flag); */
	if (!run_flag) {
		return 0;
	}
	ASSERT(!stopreq_flag);

	/* ŒZ̎sԂ𓾂 */
	exec = schedule_chkevent(microsec);
	exec2 = 0;

	do {
		/* CCPUƃTuCPU̓쑬x䗦߂ */
		cycle = schedule_get_cycle();
#if XM7_VER == 1
		if (lowspeed_submode) {
			tmp = (sub_speed_low * speed_ratio) / 100000;
		}
		else {
			tmp = (sub_speed * speed_ratio) / 100000;
		}
#else
		tmp = (sub_speed * speed_ratio) / 100000;
#endif
		if (tmp < 1) {
			tmp = 1;
		}
		ratio = (tmp << 12);
		ratio /= (cycle / 10);
#ifdef JSUB
		tmp = (jsub_speed * speed_ratio) / 100000;
		if (tmp < 1) {
			tmp = 1;
		}
		ratio_js = (tmp << 12);
		ratio_js /= (cycle / 10);
#endif

		/* CPUԂɊZ */
		count = cycle;
		count *= (exec - exec2);
		count /= 10000;
		main = (WORD)count;
		sub = (WORD)((main * ratio) >> 12);
#ifdef JSUB
		jsub = (WORD)((main * ratio_js) >> 12);
#endif

		/* JE^EtO */
		maincpu.total = main_overcycles;
		subcpu.total = sub_overcycles;
#ifdef JSUB
		jsubcpu.total = jsub_overcycles;
#endif
		mmr_modify = FALSE;

		if (cycle_steal) {
			if (break_flag) {
				/* TCNX`[Au[N|Cg */

				/* s */
				while ((maincpu.total < main) && !mmr_modify) {
					if (schedule_chkbreak()) {
						stopreq_flag = TRUE;
					}
					if (stopreq_flag) {
						break;
					}

#if XM7_VER >= 2 || (XM7_VER == 1 && defined(SFDC) && defined(IFDC))
					/* CCPUs */
					if (!dma_burst_transfer) {
						maincpu_exec();
					}

					/* DMA] */
					if (dma_flag) {
						dmac_exec();
					}
#else
					/* CCPUs */
					maincpu_exec();
#endif

					/* TuCLR߂gpBUSYtO𑀍삵ꍇ */
					/* ㏈(MAGUS) */
					if (busy_CLR_count > 0) {
						if (busy_CLR_count >= 2) {
							subbusy_flag = TRUE;
						}
						busy_CLR_count = 0;
					}

					/* TuCPUs */
					limit = maincpu.total * ratio;
					if (subhalt_flag && !(subcpu.intr & INTR_HCF)) {
						/* CCPUƋ */
						subcpu.total = (WORD)(limit >> 12);
					}
					else {
						while (((DWORD)subcpu.total << 12) < limit) {
							if (schedule_chkbreak()) {
								stopreq_flag = TRUE;
							}
							if (stopreq_flag) {
								break;
							}
							subcpu_exec();
						}
					}

#ifdef JSUB
					/* {TuCPUs */
					limit = maincpu.total * ratio_js;
					while ((((DWORD)jsubcpu.total << 12) < limit) &&
#if XM7_VER >= 2
							jsub_available && jsub_enable && !jsub_haltflag &&
							(fm_model <= MODEL_FM77AV)) {
#else
							jsub_available && jsub_enable && !jsub_haltflag &&
							(fm_model != MODEL_FM8)) {
#endif
						if (schedule_chkbreak()) {
							stopreq_flag = TRUE;
						}
						if (stopreq_flag) {
							break;
						}
						jsubcpu_exec();
					}
#endif

					/* HALTvɉ (V3.1) */
					subctrl_halt_ack();
				}

				/* u[Nꍇ̏ */
				if (stopreq_flag) {
					/* execԂi߂(ĕ␳Ȃ) */
					run_flag = FALSE;
				}
			}
			else {
				/* TCNX`[Au[N|CgȂ */

				/* s */
				while ((maincpu.total < main) && !mmr_modify) {
#if XM7_VER >= 2 || (XM7_VER == 1 && defined(SFDC) && defined(IFDC))
					/* CCPUs */
					if (!dma_burst_transfer) {
						maincpu_exec();
					}

					/* DMA] */
					if (dma_flag) {
						dmac_exec();
					}
#else
					/* CCPUs */
					maincpu_exec();
#endif

					/* TuCLR߂gpBUSYtO𑀍삵ꍇ */
					/* ㏈(MAGUS) */
					if (busy_CLR_count > 0) {
						if (busy_CLR_count >= 2) {
							subbusy_flag = TRUE;
						}
						busy_CLR_count = 0;
					}

					/* TuCPUs */
					limit = maincpu.total * ratio;
					if (subhalt_flag && !(subcpu.intr & INTR_HCF)) {
						/* CCPUƋ */
						subcpu.total = (WORD)(limit >> 12);
					}
					else {
						while (((DWORD)subcpu.total << 12) < limit) {
							subcpu_exec();
						}
					}

#ifdef JSUB
					/* {TuCPUs */
					limit = maincpu.total * ratio_js;
					while ((((DWORD)jsubcpu.total << 12) < limit) &&
#if XM7_VER >= 2
						jsub_available && jsub_enable && !jsub_haltflag &&
						(fm_model <= MODEL_FM77AV)) {
#else
						jsub_available && jsub_enable && !jsub_haltflag &&
						(fm_model != MODEL_FM8)) {
#endif
						jsubcpu_exec();
					}
#endif

					/* HALTvɉ (V3.1) */
					subctrl_halt_ack();
				}
			}
		}
		else {
			if (break_flag) {
				/* TCNX`[ȂAu[N|Cg */

				/* s */
				while ((maincpu.total < main) && !mmr_modify) {
					if (schedule_chkbreak()) {
						stopreq_flag = TRUE;
					}
					if (stopreq_flag) {
						break;
					}

#if XM7_VER >= 2 || (XM7_VER == 1 && defined(SFDC) && defined(IFDC))
					/* CCPUs */
					if (!dma_burst_transfer) {
						maincpu_exec();
					}

					/* DMA] */
					if (dma_flag) {
						dmac_exec();
					}
#else
					/* CCPUs */
					maincpu_exec();
#endif

					/* TuCLR߂gpBUSYtO𑀍삵ꍇ */
					/* ㏈(MAGUS) */
					if (busy_CLR_count > 0) {
						if (busy_CLR_count >= 2) {
							subbusy_flag = TRUE;
						}
						busy_CLR_count = 0;
					}

					/* TuCPUs */
					limit = maincpu.total * ratio;
					if ((subhalt_flag && !(subcpu.intr & INTR_HCF)) ||
						(!subclock_mode && (vrama_flag && blank_flag))) {
						/* CCPUƋ */
						subcpu.total = (WORD)(limit >> 12);
					}
					else {
						while (((DWORD)subcpu.total << 12) < limit) {
							if (schedule_chkbreak()) {
								stopreq_flag = TRUE;
							}
							if (stopreq_flag) {
								break;
							}

							subcpu_exec();
							/* VRAMANZXtOON̏ꍇ */
							/* vNbN1.83{ɂ */
							if (subclock_mode && vrama_flag) {
								subcpu.total += (WORD)(subcpu.cycle * 1.86f);
							}
						}
					}

#ifdef JSUB
					/* {TuCPUs */
					limit = maincpu.total * ratio_js;
					while ((((DWORD)jsubcpu.total << 12) < limit) &&
#if XM7_VER >= 2
							jsub_available && jsub_enable && !jsub_haltflag &&
							(fm_model <= MODEL_FM77AV)) {
#else
							jsub_available && jsub_enable && !jsub_haltflag &&
							(fm_model != MODEL_FM8)) {
#endif
						if (schedule_chkbreak()) {
							stopreq_flag = TRUE;
						}
						if (stopreq_flag) {
							break;
						}
						jsubcpu_exec();
					}
#endif

					/* HALTvɉ (V3.1) */
					subctrl_halt_ack();
				}

				/* u[Nꍇ̏ */
				if (stopreq_flag) {
					/* execԂi߂(ĕ␳Ȃ) */
					run_flag = FALSE;
				}
			}
			else {
				/* TCNX`[ȂAu[N|CgȂ */

				/* s */
				while ((maincpu.total < main) && !mmr_modify) {
#if XM7_VER >= 2 || (XM7_VER == 1 && defined(SFDC) && defined(IFDC))
					/* CCPUs */
					if (!dma_burst_transfer) {
						maincpu_exec();
					}

					/* DMA] */
					if (dma_flag) {
						dmac_exec();
					}
#else
					/* CCPUs */
					maincpu_exec();
#endif

					/* TuCLR߂gpBUSYtO𑀍삵ꍇ */
					/* ㏈(MAGUS) */
					if (busy_CLR_count > 0) {
						if (busy_CLR_count >= 2) {
							subbusy_flag = TRUE;
						}
						busy_CLR_count = 0;
					}

					/* TuCPUs */
					limit = maincpu.total * ratio;
					if ((subhalt_flag && !(subcpu.intr & INTR_HCF)) ||
						(!subclock_mode && (vrama_flag && blank_flag))) {
						/* CCPUƋ */
						subcpu.total = (WORD)(limit >> 12);
					}
					else {
						while (((DWORD)subcpu.total << 12) < limit) {
							subcpu_exec();
							/* VRAMANZXtOON̏ꍇ */
							/* vNbN1.38{ɂ */
							if (subclock_mode && vrama_flag) {
								subcpu.total += (WORD)(subcpu.cycle * 1.86f);
							}
						}
					}

#ifdef JSUB
					/* {TuCPUs */
					limit = maincpu.total * ratio_js;
					while ((((DWORD)jsubcpu.total << 12) < limit) &&
#if XM7_VER >= 2
						jsub_available && jsub_enable && !jsub_haltflag &&
						(fm_model <= MODEL_FM77AV)) {
#else
						jsub_available && jsub_enable && !jsub_haltflag &&
						(fm_model != MODEL_FM8)) {
#endif
						jsubcpu_exec();
					}
#endif

					/* HALTvɉ (V3.1) */
					subctrl_halt_ack();
				}
			}
		}

		/* MMR֘AWX^ύXꂽꍇAsԂ␳ */
		if (mmr_modify) {
			exec2 += (maincpu.total * 10000) / cycle;
			main_overcycles = 0;
			sub_overcycles = 0;
#ifdef JSUB
			jsub_overcycles = 0;
#endif
		}
	} while (mmr_modify && run_flag && (exec > exec2));

	/* I[o[TCN */
	if (maincpu.total > main) {
		main_overcycles = (WORD)(maincpu.total - main);
	}
	else {
		main_overcycles = 0;
	}
	if (subcpu.total > sub) {
		sub_overcycles = (WORD)(subcpu.total - sub);
	}
	else {
		sub_overcycles = 0;
	}
#ifdef JSUB
	if (jsubcpu.total > jsub) {
		jsub_overcycles = (WORD)(jsubcpu.total - jsub);
	}
	else {
		jsub_overcycles = 0;
	}
#endif

	/* Cxg */
	maincpu.total = 0;
	subcpu.total = 0;
#ifdef JSUB
	jsubcpu.total = 0;
#endif
	schedule_doevent(exec);
	vmtime += exec;

	return exec;
}

/*-[ t@CI/O ]----------------------------------------------------------*/

/*
 *	XPW[
 *	Z[u
 */
BOOL FASTCALL schedule_save(int fileh)
{
	int i;

	LOG0(Normal, "Z[u");

	if (!file_bool_write(fileh, run_flag)) {
		return FALSE;
	}

	if (!file_bool_write(fileh, stopreq_flag)) {
		return FALSE;
	}

	/* Ver901g */
	if (!file_byte_write(fileh, BREAKP_MAXNUM)) {
		return FALSE;
	}

	/* u[N|Cg */
	for (i=0; i<BREAKP_MAXNUM; i++) {
		if (!file_byte_write(fileh, (BYTE)breakp[i].flag)) {
			return FALSE;
		}
		if (!file_byte_write(fileh, (BYTE)breakp[i].cpu)) {
			return FALSE;
		}
		if (!file_word_write(fileh, breakp[i].addr)) {
			return FALSE;
		}
	}

	/* Cxg */

	/* Ver9g */
	if (!file_byte_write(fileh, EVENT_MAXNUM)) {
		return FALSE;
	}

	for (i=0; i<EVENT_MAXNUM; i++) {
		/* R[obNȊOۑ */
		if (!file_byte_write(fileh, (BYTE)event[i].flag)) {
			return FALSE;
		}
		if (!file_dword_write(fileh, event[i].current)) {
			return FALSE;
		}
		if (!file_dword_write(fileh, event[i].reload)) {
			return FALSE;
		}
	}

	/* Ver9.05/7.05g */
	if (!file_word_write(fileh, main_overcycles)) {
		return FALSE;
	}
	if (!file_word_write(fileh, sub_overcycles)) {
		return FALSE;
	}

#ifdef JSUB
	if (!file_word_write(fileh, jsub_overcycles)) {
		return FALSE;
	}
#endif

	/* Ver9.16/7.16/3.06g */
	if (!file_bool_write(fileh, cycle_steal)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, subclock_mode)) {
		return FALSE;
	}
#if XM7_VER == 1
	if (!file_bool_write(fileh, cycle_steal_mode)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, cycle_steal_lock)) {
		return FALSE;
	}
#endif
	if (!file_dword_write(fileh, main_speed)) {
		return FALSE;
	}
	if (!file_dword_write(fileh, mmr_speed)) {
		return FALSE;
	}
#if XM7_VER >= 2
	if (!file_dword_write(fileh, fmmr_speed)) {
		return FALSE;
	}
#endif
	if (!file_dword_write(fileh, sub_speed)) {
		return FALSE;
	}
#if XM7_VER == 1
	if (!file_dword_write(fileh, main_speed_low)) {
		return FALSE;
	}
	if (!file_dword_write(fileh, sub_speed_low)) {
		return FALSE;
	}
#endif
#ifdef JSUB
	if (!file_dword_write(fileh, jsub_speed)) {
		return FALSE;
	}
#endif

	return TRUE;
}

/*
 *	XPW[
 *	[h
 */
#if XM7_VER == 1
BOOL FASTCALL schedule_load(int fileh, int ver)
#elif defined(JSUB)
BOOL FASTCALL schedule_load(int fileh, int ver, BOOL old, BOOL jsub_flag)
#else
BOOL FASTCALL schedule_load(int fileh, int ver, BOOL old)
#endif
{
	int i;
	BYTE tmp;
	BYTE MAXNUM;
#if XM7_VER == 1 && defined(SFDC)
	DWORD dwtmp;
#endif

	/* o[W`FbN */
	if (ver < 200) {
		return FALSE;
	}

	LOG0(Normal, "[h");

	if (!file_bool_read(fileh, &run_flag)) {
		return FALSE;
	}

	if (!file_bool_read(fileh, &stopreq_flag)) {
		return FALSE;
	}

	/* u[N|Cg */
	/* Ver9g */
#if XM7_VER >= 3
	if ((ver >= 901) || ((ver >= 701) && (ver <= 799))) {
#elif XM7_VER >= 2
	if (ver >= 701) {
#else
	if (ver >= 300) {
#endif
		if (!file_byte_read(fileh, &MAXNUM)) {
			return FALSE;
		}
	}
#if XM7_VER >= 2
	else {
		/* V1.1ł͎gȂ */
		MAXNUM = BREAKP_MAXNUM_OLD;
	}
#endif

	/* 񏉊 */
	memset(breakp, 0, sizeof(breakp));

	for (i=0; i<MAXNUM; i++) {
		if (!file_byte_read(fileh, &tmp)) {
			return FALSE;
		}
		breakp[i].flag = (int)tmp;
		if (!file_byte_read(fileh, &tmp)) {
			return FALSE;
		}
		breakp[i].cpu = (int)tmp;
		if (!file_word_read(fileh, &breakp[i].addr)) {
			return FALSE;
		}
	}
	break_flag = FALSE;
	for (i=0; i<MAXNUM; i++) {
		if (breakp[i].flag != BREAKP_NOTUSE) {
			break_flag = TRUE;
		}
	}

	/* Cxg */
	/* Ver9g */
#if XM7_VER >= 3
	if ((ver >= 900) || ((ver >= 700) && (ver <= 799))) {
#elif XM7_VER >= 2
	if (ver >= 700) {
#else
	if (ver >= 300) {
#endif
		if (!file_byte_read(fileh, &MAXNUM)) {
			return FALSE;
		}
	}
#if XM7_VER >= 2
	else {
		if (old) {
			MAXNUM = EVENT_MAXNUM_L30;
		}
		else {
			MAXNUM = EVENT_MAXNUM_L31;
		}
	}
#endif

	for (i=0; i<MAXNUM; i++) {
#if XM7_VER == 1 && defined(SFDC)
		/*	SFDC֘A𖢎gpCxgɊ蓖Ă	*/
		/*	\Cxg΍					*/
		if (i >= EVENT_MAXNUM) {
			if (!file_byte_read(fileh, &tmp)) {
				return FALSE;
			}
			if (!file_dword_read(fileh, &dwtmp)) {
				return FALSE;
			}
			if (!file_dword_read(fileh, &dwtmp)) {
				return FALSE;
			}
			continue;
		}
#endif

		/* R[obNȊOݒ */
		if (!file_byte_read(fileh, &tmp)) {
			return FALSE;
		}
		event[i].flag = (int)tmp;
		if (!file_dword_read(fileh, &event[i].current)) {
			return FALSE;
		}
		if (!file_dword_read(fileh, &event[i].reload)) {
			return FALSE;
		}
	}

	/* Ver9.05/Ver7.05g */
#if XM7_VER >= 3
	if ((ver >= 905) || ((ver >= 705) && (ver <= 799))) {
#elif XM7_VER >= 2
	if (ver >= 705) {
#else
	if (ver >= 300) {
#endif
		if (!file_word_read(fileh, &main_overcycles)) {
			return FALSE;
		}
		if (!file_word_read(fileh, &sub_overcycles)) {
			return FALSE;
		}
#ifdef JSUB
#if XM7_VER >= 2
		if ((ver >= 1200) || jsub_flag) {
			if (!file_word_read(fileh, &jsub_overcycles)) {
				return FALSE;
			}
		}
		else {
			jsub_overcycles = 0;
		}
#else
		if (!file_word_read(fileh, &jsub_overcycles)) {
			return FALSE;
		}
#endif
#endif
	}
	else {
		main_overcycles = 0;
		sub_overcycles = 0;
#ifdef JSUB
		jsub_overcycles = 0;
#endif
	}

	/* Ver9.16/7.16/3.06g */
#if XM7_VER >= 3
	if ((ver >= 916) || ((ver >= 716) && (ver <= 799))) {
#elif XM7_VER >= 2
	if (ver >= 716) {
#else
	if ((ver >= 404) || ((ver >= 306) && (ver <= 399))) {
#endif
		if (!file_bool_read(fileh, &cycle_steal)) {
			return FALSE;
		}
#if XM7_VER >= 3
		if (((ver >= 1301) && (ver <= 1399)) ||
			((ver >= 1501) && (ver <= 1599)) ||
			((ver >= 723) && (ver <= 799)) ||
			((ver >= 923) && (ver <= 999))) {
#elif XM7_VER >= 2
		if (((ver >= 1301) && (ver <= 1399)) ||
			((ver >= 723) && (ver <= 799))) {
#else
		if (((ver >= 1101) && (ver <= 1399)) ||
			((ver >= 313) && (ver <= 399))) {
#endif
			if (!file_bool_read(fileh, &subclock_mode)) {
				return FALSE;
			}
			else {
				if (cycle_steal) {
					subclock_mode = FALSE;
				}
			}
		}
#if XM7_VER == 1
	}
	if (ver >= 1007) {
		if (!file_bool_read(fileh, &cycle_steal_mode)) {
			return FALSE;
		}
		if (!file_bool_read(fileh, &cycle_steal_lock)) {
			return FALSE;
		}
	}
	else {
		cycle_steal_mode = cycle_steal;
		cycle_steal_lock = FALSE;
	}

	if ((ver >= 407) || ((ver >= 306) && (ver <= 399))) {
#endif	/* XM7_VER == 1 */
		if (!file_dword_read(fileh, &main_speed)) {
			return FALSE;
		}
		if (!file_dword_read(fileh, &mmr_speed)) {
			return FALSE;
		}
#if XM7_VER >= 2
#if XM7_VER >= 3
		if (ver >= 916) {
#else
		if (ver >= 1200) {
#endif
			if (!file_dword_read(fileh, &fmmr_speed)) {
				return FALSE;
			}
		}
#endif
		if (!file_dword_read(fileh, &sub_speed)) {
			return FALSE;
		}
#if XM7_VER == 1
		if (!file_dword_read(fileh, &main_speed_low)) {
			return FALSE;
		}
		if (!file_dword_read(fileh, &sub_speed_low)) {
			return FALSE;
		}
#endif
#ifdef JSUB
#if XM7_VER >= 2
		if ((ver >= 1200) || jsub_flag) {
			if (!file_dword_read(fileh, &jsub_speed)) {
				return FALSE;
			}
		}
		else {
			jsub_speed = JSUBCYCLES * 10;
		}
#else
		if (!file_dword_read(fileh, &jsub_speed)) {
			return FALSE;
		}
#endif
#endif
	}

	/* sx䗦 */
	speed_ratio = 10000;

#if XM7_VER == 1 && defined(SFDC)
	/* ]Cxg폜 */
	if (MAXNUM < EVENT_MAXNUM) {
		for (i=MAXNUM; i<EVENT_MAXNUM; i++) {
			schedule_delevent(i);
		}
	}

	/* SFDCCxg */
	if (ver <= 401) {
		if (event[19].flag != EVENT_NOTUSE) {
			event[EVENT_FDC_NMI].flag    = event[19].flag;
			event[EVENT_FDC_NMI].current = event[19].current;
			event[EVENT_FDC_NMI].reload  = event[19].reload;
			event[19].flag    = EVENT_NOTUSE;
			event[19].current = 0;
			event[19].reload  = 0;
		}
	}
	else if (ver <= 410) {
		if (event[25].flag != EVENT_NOTUSE) {
			event[EVENT_SFDD_SEEK].flag    = event[25].flag;
			event[EVENT_SFDD_SEEK].current = event[25].current;
			event[EVENT_SFDD_SEEK].reload  = event[25].reload;
			event[25].flag    = EVENT_NOTUSE;
			event[25].current = 0;
			event[25].reload  = 0;
		}
	}
	else if (ver <= 1000) {
		if (event[24].flag != EVENT_NOTUSE) {
			event[EVENT_SFDC_M].flag    = event[24].flag;
			event[EVENT_SFDC_M].current = event[24].current;
			event[EVENT_SFDC_M].reload  = event[24].reload;
			event[24].flag    = EVENT_NOTUSE;
			event[24].current = 0;
			event[24].reload  = 0;
		}
		if (event[25].flag != EVENT_NOTUSE) {
			event[EVENT_SFDC_L].flag    = event[25].flag;
			event[EVENT_SFDC_L].current = event[25].current;
			event[EVENT_SFDC_L].reload  = event[25].reload;
			event[25].flag    = EVENT_NOTUSE;
			event[25].current = 0;
			event[25].reload  = 0;
		}
		if (event[26].flag != EVENT_NOTUSE) {
			event[EVENT_SFDD_SEEK].flag    = event[26].flag;
			event[EVENT_SFDD_SEEK].current = event[26].current;
			event[EVENT_SFDD_SEEK].reload  = event[26].reload;
			event[26].flag    = EVENT_NOTUSE;
			event[26].current = 0;
			event[26].reload  = 0;
		}
	}
#endif	/* XM7_VER == 1 && defined(SFDC) */

#if XM7_VER == 1 && defined(BUBBLE)
	/* ouCxg */
	if (ver < 400) {
		if (event[27].flag != EVENT_NOTUSE) {
			event[EVENT_BMC_M].flag    = event[27].flag;
			event[EVENT_BMC_M].current = event[27].current;
			event[EVENT_BMC_M].reload  = event[27].reload;
			event[27].flag    = EVENT_NOTUSE;
			event[27].current = 0;
			event[27].reload  = 0;
		}
	}
#endif	/* XM7_VER == 1 && defined(BUBBLE) */

#if XM7_VER == 1 && defined(FDDSND)
	if (ver < 400) {
	/* XM7+ FDD[^Cxg */
		if (event[28].flag != EVENT_NOTUSE) {
			event[EVENT_FDD_MOTOR].flag    = event[28].flag;
			event[EVENT_FDD_MOTOR].current = event[28].current;
			event[EVENT_FDD_MOTOR].reload  = event[28].reload;
			event[28].flag    = EVENT_NOTUSE;
			event[28].current = 0;
			event[28].reload  = 0;
		}
	}
#endif	/* XM7_VER == 1 && defined(FDDSND) */

	return TRUE;
}
