/*
 *	FM-7 EMULATOR "XM7"
 *
 *	Copyright (C) 1999-2020 ohD(Twitter:@xm6_original)
 *	Copyright (C) 2001-2020 Ryu Takegami (Twitter:@RyuTakegami)
 *	Copyright (C) 2010-2018 Toma
 *
 *	[ VXeǗ ]
 */

#include <string.h>
#include <stdlib.h>
#include "xm7.h"
#include "display.h"
#include "ttlpalet.h"
#include "subctrl.h"
#include "keyboard.h"
#include "fdc.h"
#include "mainetc.h"
#include "multipag.h"
#include "kanji.h"
#include "tapelp.h"
#include "display.h"
#include "opn.h"
#include "mmr.h"
#include "aluline.h"
#include "apalet.h"
#include "rtc.h"
#include "mouse.h"
#include "rs232c.h"
#include "jsubsys.h"
#include "bubble.h"
#ifdef MIDI
#include "midi.h"
#endif
#if XM7_VER >= 2
#include "jcard.h"
#endif
#if XM7_VER >= 2 || (XM7_VER == 1 && defined(SFDC))
#include "dmac.h"
#endif
#include "ptm.h"
#include "device.h"
#include "event.h"

/*
 *	O[o [N
 */
int fm7_ver;							/* n[hEFAo[W */
int boot_mode;							/* N[h BASIC/DOS */
BYTE fm_model;							/* n[hEFAf */
#if XM7_VER == 1
BOOL lowspeed_mode;						/* NbN[h */
BOOL lowspeed_submode;					/* TuCPUNbN[h */
BOOL available_fm8roms;					/* FM-8 ROMgp\tO */
BOOL available_fm7roms;					/* FM-7 ROMgp\tO */
BOOL available_bblboot8;				/* FM-8 BOOTROM(BUBBLE)gp\tO */
BOOL available_sfdboot8;				/* FM-8 BOOTROM(8"DOS)gp\tO */
BOOL available_dbgboot8;				/* FM-8 BOOTROM(DEBUG)gp\tO */
BOOL available_1mbboot;					/* FM-77 BOOTROM(1MB DOS)gp\tO */
BOOL available_bblboot;					/* FM-77 BOOTROM(BUBBLE)gp\tO */
#endif

BOOL hotreset_flag;						/* CzbgZbgtO */
BOOL reset_flag;						/* VXeZbgtO */
BYTE fetch_op;							/* OɃtFb` */

/*
 *	Xe[gt@Cwb_
 */
#if XM7_VER >= 3
const char *state_header = "XM7rVM STATE F02";
#elif XM7_VER >= 2
const char *state_header = "XM7rVM STATE D02";
#else
const char *state_header = "XM7rVM STATE B02";
#endif

/*
 *	ID
 */
#define	ID	MAKEID('S', 'Y', 'S', ' ')
//#define	SYSTEM_LOG

/*
 *	VXe
 *	
 */
BOOL FASTCALL system_init(void)
{
	/* [hݒ */
#if XM7_VER >= 3
	fm7_ver = 3;					/* FM77AV40EXɐݒ */
	fm_model = MODEL_AV40EX;
#elif XM7_VER >= 2
	fm7_ver = 2;					/* FM77AVɐݒ */
	fm_model = MODEL_FM77AV;
#else
	fm7_ver = 1;					/* FM-77+192KB RAMɐݒ */
	fm_model = MODEL_FM77;
	lowspeed_mode = FALSE;
	lowspeed_submode = FALSE;
	available_fm8roms = TRUE;
	available_fm7roms = TRUE;
	available_bblboot8 = TRUE;
	available_sfdboot8 = TRUE;
	available_dbgboot8 = TRUE;
	available_1mbboot = TRUE;
	available_bblboot = TRUE;
#endif
	boot_mode = BOOT_BASIC;			/* BASIC MODE */

	hotreset_flag = FALSE;
	reset_flag = FALSE;

	/* XPW[AoX */
	if (!schedule_init()) {
		return FALSE;
	}
	if (!mmr_init()) {
		return FALSE;
	}

	/* ACPU */
	if (!mainmem_init()) {
		return FALSE;
	}
	if (!submem_init()) {
		return FALSE;
	}
	if (!maincpu_init()) {
		return FALSE;
	}
	if (!subcpu_init()) {
		return FALSE;
	}
#ifdef JSUB
	if (!jsubsys_init()) {
		return FALSE;
	}
#endif

	/* ̑foCX */
	if (!display_init()) {
		return FALSE;
	}
	if (!ttlpalet_init()) {
		return FALSE;
	}
	if (!subctrl_init()) {
		return FALSE;
	}
	if (!keyboard_init()) {
		return FALSE;
	}
	if (!fdc_init()) {
		return FALSE;
	}
	if (!mainetc_init()) {
		return FALSE;
	}
	if (!multipag_init()) {
		return FALSE;
	}
	if (!kanji_init()) {
		return FALSE;
	}
	if (!tapelp_init()) {
		return FALSE;
	}
	if (!opn_init()) {
		return FALSE;
	}
	if (!whg_init()) {
		return FALSE;
	}
	if (!thg_init()) {
		return FALSE;
	}
	if (!psg_init()) {
		return FALSE;
	}
#if XM7_VER >= 2
	if (!aluline_init()) {
		return FALSE;
	}
	if (!apalet_init()) {
		return FALSE;
	}
	if (!rtc_init()) {
		return FALSE;
	}
#endif
#ifdef MIDI
	if (!midi_init()) {
		return FALSE;
	}
#endif
#if XM7_VER >= 2
	if (!jcard_init()) {
		return FALSE;
	}
#endif
#if XM7_VER >= 2 || (XM7_VER == 1 && defined(SFDC) && defined(IFDC))
	if (!dmac_init()) {
		return FALSE;
	}
#endif
#ifdef MOUSE
	if (!mos_init()) {
		return FALSE;
	}
	if (!ptm_init()) {
		return FALSE;
	}
#endif
#if (XM7_VER == 1) && defined(BUBBLE)
	if (!bmc_init()) {
		return FALSE;
	}
#endif

	return TRUE;
}

/*
 *	VXe
 *	N[Abv
 */
void FASTCALL system_cleanup(void)
{
	/* ̑foCX */
#if (XM7_VER == 1) && defined(BUBBLE)
	bmc_cleanup();
#endif
#ifdef MOUSE
	ptm_cleanup();
	mos_cleanup();
#endif
#if XM7_VER >= 2 || (XM7_VER == 1 && defined(SFDC) && defined(IFDC))
	dmac_cleanup();
#endif
#if XM7_VER >= 2
	jcard_cleanup();
#endif
#ifdef MIDI
	midi_cleanup();
#endif
#if XM7_VER >= 2
	rtc_cleanup();
	apalet_cleanup();
	aluline_cleanup();
#endif
	psg_cleanup();
	thg_cleanup();
	whg_cleanup();
	opn_cleanup();
	tapelp_cleanup();
	kanji_cleanup();
	multipag_cleanup();
	mainetc_cleanup();
	fdc_cleanup();
	keyboard_cleanup();
	subctrl_cleanup();
	ttlpalet_cleanup();
	display_cleanup();

	/* ACPU */
#ifdef JSUB
	jsubsys_cleanup();
#endif
	subcpu_cleanup();
	maincpu_cleanup();
	submem_cleanup();
	mainmem_cleanup();

	/* XPW[AoX */
	mmr_cleanup();
	schedule_cleanup();
}

/*
 *	VXe
 *	Zbg
 */
void FASTCALL system_reset(void)
{
	LOG0(Normal, "Zbg");

	/* XPW[AoX */
	schedule_reset();
	mmr_reset();

	/* ̑foCX */
	display_reset();
	ttlpalet_reset();
	subctrl_reset();
	keyboard_reset();
	fdc_reset();
	mainetc_reset();
	multipag_reset();
	kanji_reset();
	tapelp_reset();
	opn_reset();
	whg_reset();
	thg_reset();
	psg_reset();
#if XM7_VER >= 2
	aluline_reset();
	apalet_reset();
	rtc_reset();
#endif
#ifdef MIDI
	midi_reset();
#endif
#ifdef RSC
	rs232c_reset();
#endif
#if XM7_VER >= 2
	jcard_reset();
#endif
#if XM7_VER >= 2 || (XM7_VER == 1 && defined(SFDC) && defined(IFDC))
	dmac_reset();
#endif
#ifdef MOUSE
	mos_reset();
	ptm_reset();
#endif
#if (XM7_VER == 1) && defined(BUBBLE)
	bmc_reset();
#endif

	/* ACPU */
	mainmem_reset();
	submem_reset();
	maincpu_reset();
	subcpu_reset();
#ifdef JSUB
	jsubsys_reset();
#endif

	/* ʍĕ` */
	display_setpointer(FALSE);
	display_notify();

	/* VXeZbgtOݒ */
	reset_flag = TRUE;

	/* tFb`ߏ */
	fetch_op = 0;
}

/*
 *	VXe
 *	zbgZbg
 */
void FASTCALL system_hotreset(void)
{
#if XM7_VER == 1
	BOOL flag;
#endif

	/* TWRLȂN[hDOSɐ؂ւ */
#if XM7_VER >= 2
	if (twr_flag) {
		boot_mode = BOOT_DOS;
	}
#else
	flag = twr_flag;
#endif

	/* VXeZbg */
	system_reset();

	/* BREAKL[Ԃɂ */
	break_flag = TRUE;
	maincpu_firq();

	/* zbgZbgtOL */
	hotreset_flag = TRUE;

#if XM7_VER == 1
	/* ZbgOTWRL痠RAMI */
	if (flag) {
		basicrom_en = FALSE;
	}
#endif
}

/*
 *	VXe
 *	TAB+Zbg
 */
#if XM7_VER >= 2
void FASTCALL system_tabreset(void)
{
	/* VXeZbg */
	system_reset();

	/* TABL[R[hs */
	key_fm7 = 0x09;
}
#endif

/*
 *	VXe
 *	t@CZ[u
 */
BOOL FASTCALL system_save(char *filename)
{
	int fileh;
	BOOL flag;
	ASSERT(filename);

	LOG0(Normal, "Z[u");

	/* t@CI[v */
	fileh = file_open(filename, OPEN_W);
	if (fileh == -1) {
		return FALSE;
	}

	/* tO */
	flag = TRUE;

	/* wb_Z[u */
	if (!file_write(fileh, (BYTE *)state_header, 16)) {
		flag = FALSE;
	}

	/* VXe[N */
	if (!file_word_write(fileh, (WORD)fm7_ver)) {
		return FALSE;
	}
	if (!file_word_write(fileh, (WORD)boot_mode)) {
		return FALSE;
	}
	if (!file_byte_write(fileh, fm_model)) {
		return FALSE;
	}
#if XM7_VER == 1
	if (!file_bool_write(fileh, lowspeed_mode)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, available_fm8roms)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, available_fm7roms)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, available_bblboot8)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, available_sfdboot8)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, available_dbgboot8)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, available_1mbboot)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, available_bblboot)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, TRUE)) {
		return FALSE;
	}
	if (!file_bool_write(fileh, lowspeed_submode)) {
		return FALSE;
	}
#endif

	/* ԂɌĂяo */
	if (!mainmem_save(fileh)) {
		flag = FALSE;
	}
	if (!submem_save(fileh)) {
		flag = FALSE;
	}
	if (!maincpu_save(fileh)) {
		flag = FALSE;
	}
	if (!subcpu_save(fileh)) {
		flag = FALSE;
	}
	if (!schedule_save(fileh)) {
		flag = FALSE;
	}
	if (!display_save(fileh)) {
		flag = FALSE;
	}
	if (!ttlpalet_save(fileh)) {
		flag = FALSE;
	}
	if (!subctrl_save(fileh)) {
		flag = FALSE;
	}
	if (!keyboard_save(fileh)) {
		flag = FALSE;
	}
	if (!fdc_save(fileh)) {
		flag = FALSE;
	}
	if (!mainetc_save(fileh)) {
		flag = FALSE;
	}
	if (!multipag_save(fileh)) {
		flag = FALSE;
	}
	if (!kanji_save(fileh)) {
		flag = FALSE;
	}
	if (!tapelp_save(fileh)) {
		flag = FALSE;
	}
	if (!opn_save(fileh)) {
		flag = FALSE;
	}
	if (!mmr_save(fileh)) {
		flag = FALSE;
	}
#if XM7_VER >= 2
	if (!aluline_save(fileh)) {
		flag = FALSE;
	}
	if (!rtc_save(fileh)) {
		flag = FALSE;
	}
	if (!apalet_save(fileh)) {
		flag = FALSE;
	}
#endif
	if (!whg_save(fileh)) {
		flag = FALSE;
	}
	if (!thg_save(fileh)) {
		flag = FALSE;
	}
	if (!psg_save(fileh)) {
		flag = FALSE;
	}
#if XM7_VER >= 2
	if (!jcard_save(fileh)) {
		flag = FALSE;
	}
#endif
#if XM7_VER >= 2 || (XM7_VER == 1 && defined(SFDC) && defined(IFDC))
	if (!dmac_save(fileh)) {
		flag = FALSE;
	}
#endif
#ifdef MOUSE
	if (!mos_save(fileh)) {
		flag = FALSE;
	}
#endif
#ifdef RSC
	if (!rs232c_save(fileh)) {
		flag = FALSE;
	}
#endif
#ifdef MIDI
	if (!midi_save(fileh)) {
		flag = FALSE;
	}
#endif
#ifdef JSUB
	if (!jsubsys_save(fileh)) {
		flag = FALSE;
	}
#endif
#if (XM7_VER == 1) && defined(BUBBLE)
	if (!bmc_save(fileh)) {
		flag = FALSE;
	}
#endif
#if XM7_VER == 1 && defined(SFDC)
	if (!sfdc_save(fileh)) {
		flag = FALSE;
	}
#endif	/* XM7_VER == 1 && defined(SFDC) */
#ifdef MOUSE
	if (!ptm_save(fileh)) {
		flag = FALSE;
	}
#endif

	file_close(fileh);
	return flag;
}

/*
 *	VXe
 *	t@C[h
 */
int FASTCALL system_load(char *filename)
{
	int fileh;
	int ver;
	char header[16];
	BOOL flag;
#if XM7_VER >= 2
	BOOL old_scheduler;
	int filesize;
#endif
	WORD tmp;
#if XM7_VER == 1
	BOOL temp;
#endif
	BOOL dash_flag;
	BOOL typer_flag;
#if XM7_VER >= 2
	BOOL jsub_flag;
#if XM7_VER >= 3
	BOOL model_flag;
#endif
#endif
	char state_ver_base;

	ASSERT(filename);

	LOG0(Normal, "[h");

	/* t@CI[v */
	fileh = file_open(filename, OPEN_R);
	if (fileh == -1) {
		return STATELOAD_OPENERR;
	}

	/* tO */
	flag = TRUE;
#if XM7_VER >= 2
	old_scheduler = FALSE;
#endif
	dash_flag = FALSE;
	typer_flag = FALSE;
#if XM7_VER >= 2
	jsub_flag = FALSE;
#if XM7_VER >= 3
	model_flag = FALSE;
#endif
#endif

	/* wb_[h */
	if (!file_read(fileh, (BYTE *)header, 16)) {
		flag = FALSE;
	}
	else {
		/* dash_flag̃Xe[g */
		if ((header[3] == 0x27) || (header[3] == 'r')) {
			if (header[3] == 0x27) {
				dash_flag = TRUE;
			}
			else if (header[3] == 'r') {
				typer_flag = TRUE;
			}
			header[3] = 0x20;

#if XM7_VER >= 2
			/* Ver9.18(V3)JSUB, Ver7.18(V2)JCARD/JSUBLtO #1 */
			if (header[6] == 0x27) {
				header[6] = 0x20;
				jsub_flag = TRUE;
			}
#endif
		}
		if (memcmp(header, "XM7 VM STATE ", 13) != 0) {
			flag = FALSE;
		}
	}

	/* wb_`FbN */
	if (!flag) {
		file_close(fileh);
		return STATELOAD_HEADERR;
	}

	/* t@Co[W擾Ao[W2ȏオΏ */
	if (header[13] != 0x20) {
		if (header[13] >= 0x41) {
			state_ver_base = 0x37;
		}
		else {
			state_ver_base = 0x30;
		}
		ver = (int)	((BYTE)(header[13] - state_ver_base) * 100) +
					((BYTE)(header[14] - 0x30) * 10) +
					((BYTE)(header[15] - 0x30));
	}
	else {
		ver = (int)(BYTE)(header[15]);
		ver -= 0x30;
		ver *= 100;
	}
#if XM7_VER >= 3
	/* V3 : TypeR Ver14.09Ver15.00ɁAVer12.09Ver13.00ɓǂݑւ */
	if (!dash_flag && typer_flag && (ver == 1409)) {
		ver = 1500;
	}
	if (!dash_flag && typer_flag && (ver == 1209)) {
		ver = 1300;
	}
#elif XM7_VER >= 2
	/* V2 : TypeR Ver12.09Ver13.00ɓǂݑւ */
	if (!dash_flag && typer_flag && (ver == 1209)) {
		ver = 1300;
	}
#else
	/* V1 : TypeR Ver10.09Ver11.00ɓǂݑւ */
	if (!dash_flag && typer_flag && (ver == 1009)) {
		ver = 1100;
	}
#endif
#if XM7_VER >= 3
	/* V3 : Ver5EVer10-11̓[hłȂ */
	if ((ver < 200) || ((ver >= 300) && (ver <= 499)) ||
		((ver >= 1000) && (ver <= 1199))) {
		file_close(fileh);
		return STATELOAD_VERERR;
	}
#elif XM7_VER >= 2
	/* V2 : Ver5EVer8-11EVer14ȏ̓[hłȂ */
	if ((ver < 200) || ((ver >= 300) && (ver <= 499)) ||
		((ver >= 800) && (ver <= 1199)) || (ver >= 1400)) {
		file_close(fileh);
		return STATELOAD_VERERR;
	}
#else
	/* V1 : Ver2EVer5-9EVer12ȏ̓[hłȂ */
	if ((ver < 200) || ((ver >= 500) && (ver <= 999)) || (ver >= 1200)) {
		file_close(fileh);
		return STATELOAD_VERERR;
	}
#endif

	/* XM7Xe[gt@Č`ύXƂ̑΍ */
#if XM7_VER >= 3
	/* V3 : XM7Ver9.25ȏ̓[hłȂ(_ł݂͑Ȃ͂) */
	if (!dash_flag && !typer_flag && ((ver >= 925) ||
		((ver >= 725) && (ver <= 799)))) {
		file_close(fileh);
		return STATELOAD_VERERR;
	}
#elif XM7_VER >= 2
	/* V2 : XM7Ver7.25ȏ̓[hłȂ(_ł݂͑Ȃ͂) */
	if (!dash_flag && !typer_flag && (ver >= 725)) {
		file_close(fileh);
		return STATELOAD_VERERR;
	}
#else
	/* V1 : XM7Ver3.15ȏ̓[hłȂ(\_ł݂͑Ȃ͂) */
	if (!dash_flag && !typer_flag && (ver >= 315)) {
		file_close(fileh);
		return STATELOAD_VERERR;
	}
#endif

	/* XM7dashXe[gt@Č`ύXƂ̑΍ */
#if XM7_VER >= 3
	/* V3 : XM7dashVer14.10`14.99̓[hłȂ */
	if (dash_flag && !typer_flag && ((ver >= 1410) && (ver <= 1499)) ||
		((ver >= 1210) && (ver <= 1299))) {
		file_close(fileh);
		return STATELOAD_VERERR;
	}
#elif XM7_VER >= 2
	/* V2 : XM7Ver12.10`12.99̓[hłȂ */
	if (dash_flag && !typer_flag && ((ver >= 1210) && (ver <= 1299))) {
		file_close(fileh);
		return STATELOAD_VERERR;
	}
#else
	/* V1 : XM7Ver10.10`10.99̓[hłȂ */
	if (dash_flag && !typer_flag && ((ver >= 1010) && (ver <= 1099))){
		file_close(fileh);
		return STATELOAD_VERERR;
	}
#endif

#if XM7_VER >= 2
#if XM7_VER >= 3
	if ((dash_flag || typer_flag) &&
		((ver >= 918) || ((ver >= 718) && (ver <= 799)))) {
		if (ver == 918) {
			/* Ver9.18(V3)fm_model/DMAC_AV40LtO */
			model_flag = !jsub_flag;
		}
#else
	if ((dash_flag || typer_flag) && (ver >= 718)) {
#endif
		/* Ver9.18(V3)JSUB, Ver7.18(V2)JCARD/JSUBLtO #2 */
		jsub_flag = TRUE;
	}
#endif
	if (state_header[13] >= 0x41) {
		state_ver_base = 0x37;
	}
	else {
		state_ver_base = 0x30;
	}
	if (ver >	(int)((BYTE)(state_header[13] - state_ver_base) * 100) +
					 ((BYTE)(state_header[14] - 0x30) * 10) +
					 ((BYTE)(state_header[15] - 0x30))) {
		file_close(fileh);
		return STATELOAD_VERERR;
	}

#if XM7_VER >= 2
	/* V3.0L30/V2.5L20ȑÕXe[gt@CɑΏ */
	filesize = file_getsize(fileh);
	if (ver <= 500) {
		old_scheduler = TRUE;
	}
	else {
#if XM7_VER >= 3
		if ((filesize == 454758) ||		/* XM7 V3.0 (RAM) */
			(filesize == 1241190) ||	/* XM7 V3.0 (RAML) */
			(filesize == 258030)) {		/* XM7 V2 */
			old_scheduler = TRUE;
		}
#else
		if (filesize == 258030) {		/* XM7 V2 */
			old_scheduler = TRUE;
		}
#endif
	}
#endif

	/* VXe[N */
	if (!file_word_read(fileh, &tmp)) {
		return FALSE;
	}
	fm7_ver = (int)tmp;
	if (!file_word_read(fileh, &tmp)) {
		return FALSE;
	}
	boot_mode = (int)tmp;
#if XM7_VER == 1 && defined(BUBBLE)
	if ((ver >= 307) && (ver <= 399)) {
		if (boot_mode == 2) {
			boot_mode = BOOT_BUBBLE;
		}
	}
#ifdef BUBBLE128
	mainmem_bubblemode_boot();
#endif
#endif	/* XM7_VER == 1 && defined(BUBBLE) */
#if XM7_VER >= 3
	if (((ver >= 719) && (ver <= 799)) || (ver >= 919) ||
		((dash_flag || typer_flag) && model_flag)) {
#elif XM7_VER >= 2
	if (ver >= 719) {
#else
	{
#endif
		if (!file_byte_read(fileh, &fm_model)) {
			return FALSE;
		}
	}
#if XM7_VER >= 2
#if XM7_VER >= 3
	if ((ver <= 999) && ((ver != 918) || !dash_flag ||
		!typer_flag || !model_flag)) {
#else
	if (ver <= 799) {
#endif
		switch (fm7_ver) {
			case 1:
				/* FM-7 */
				fm_model = MODEL_FM7;
				break;
			case 2:
				/* FM77AV */
				fm_model = MODEL_FM77AV;
				break;
#if XM7_VER >= 3
			case 3:
				/* FM77AV40EX */
				fm_model = MODEL_AV40EX;
				break;
#endif
		}
	}
#endif
#if XM7_VER == 1
	if (!file_bool_read(fileh, &lowspeed_mode)) {
		return FALSE;
	}

	/* Z[uɗLROMf[^[hɗLłȂꍇG[ */
	if (!file_bool_read(fileh, &temp)) {
		return FALSE;
	}
	if (!available_fm8roms && temp) {
		return FALSE;
	}
	if (!file_bool_read(fileh, &temp)) {
		return FALSE;
	}
	if (!available_fm7roms && temp) {
		return FALSE;
	}

	/* XM7 V1.2 - Ver4g BOOT ROM֘A */
	if (ver >= 400) {
		if (!file_bool_read(fileh, &temp)) {
			return FALSE;
		}
		if (!available_bblboot8 && temp) {
			return FALSE;
		}
		if (!file_bool_read(fileh, &temp)) {
			return FALSE;
		}
#ifdef IFDC
		if (!available_sfdboot8 && temp) {
			return FALSE;
		}
#endif
		if (ver >= 408) {
			if (!file_bool_read(fileh, &temp)) {
				return FALSE;
			}
			if (!available_dbgboot8 && temp) {
				return FALSE;
			}
		}
		if (!file_bool_read(fileh, &temp)) {
			return FALSE;
		}
		if (!available_1mbboot && temp) {
			return FALSE;
		}
		if (ver >= 404) {
			if (!file_bool_read(fileh, &temp)) {
				return FALSE;
			}
			if (!available_bblboot && temp) {
				return FALSE;
			}
		}
		if (!file_bool_read(fileh, &temp)) {
			return FALSE;
		}
		if (ver >= 1002) {
			if (!file_bool_read(fileh, &lowspeed_submode)) {
				return FALSE;
			}
		}
	}
#endif

	/* ԂɌĂяo */
	if (!mainmem_load(fileh, ver)) {
		flag = FALSE;
	}
	if (!submem_load(fileh, ver)) {
		flag = FALSE;
	}
	if (!maincpu_load(fileh, ver)) {
		flag = FALSE;
	}
	if (!subcpu_load(fileh, ver)) {
		flag = FALSE;
	}
#if XM7_VER == 1
	if (!schedule_load(fileh, ver)) {
		flag = FALSE;
	}
#elif defined(JSUB)
	if (!schedule_load(fileh, ver, old_scheduler, jsub_flag)) {
		flag = FALSE;
	}
#else
	if (!schedule_load(fileh, ver, old_scheduler)) {
		flag = FALSE;
	}
#endif
	if (!display_load(fileh, ver)) {
		flag = FALSE;
	}
	if (!ttlpalet_load(fileh, ver)) {
		flag = FALSE;
	}
	if (!subctrl_load(fileh, ver)) {
		flag = FALSE;
	}
	if (!keyboard_load(fileh, ver)) {
		flag = FALSE;
	}
	if (!fdc_load(fileh, ver)) {
		flag = FALSE;
	}
	if (!mainetc_load(fileh, ver)) {
		flag = FALSE;
	}
	if (!multipag_load(fileh, ver)) {
		flag = FALSE;
	}
	if (!kanji_load(fileh, ver)) {
		flag = FALSE;
	}
	if (!tapelp_load(fileh, ver)) {
		flag = FALSE;
	}
	if (!opn_load(fileh, ver)) {
		flag = FALSE;
	}
	if (!mmr_load(fileh, ver)) {
		flag = FALSE;
	}
#if XM7_VER >= 2
	if (!aluline_load(fileh, ver)) {
		flag = FALSE;
	}
	if (!rtc_load(fileh, ver)) {
		flag = FALSE;
	}
	if (!apalet_load(fileh, ver)) {
		flag = FALSE;
	}
#endif
	if (!whg_load(fileh, ver)) {
		flag = FALSE;
	}
	if (!thg_load(fileh, ver)) {
		flag = FALSE;
	}
	if (!psg_load(fileh, ver)) {
		flag = FALSE;
	}
#if XM7_VER >= 2
	if (!jcard_load(fileh, ver, jsub_flag)) {
		flag = FALSE;
	}
#endif
#if XM7_VER >= 3
	if (!dmac_load(fileh, ver, model_flag)) {
		flag = FALSE;
	}
#elif XM7_VER >= 2 || (XM7_VER == 1 && defined(SFDC) && defined(IFDC))
	if (!dmac_load(fileh, ver)) {
		flag = FALSE;
	}
#endif	/* XM7_VER >= 3 */
#ifdef MOUSE
	if (!mos_load(fileh, ver)) {
		flag = FALSE;
	}
#endif
#ifdef RSC
	if (!rs232c_load(fileh, ver)) {
		flag = FALSE;
	}
#endif
#ifdef MIDI
	if (!midi_load(fileh, ver)) {
		flag = FALSE;
	}
#endif
#ifdef JSUB
#if XM7_VER >= 2
	if (!jsubsys_load(fileh, ver, jsub_flag)) {
		flag = FALSE;
	}
#else
	if (!jsubsys_load(fileh, ver)) {
		flag = FALSE;
	}
#endif
#endif
#if (XM7_VER == 1) && defined(BUBBLE)
	if (!bmc_load(fileh, ver)) {
		flag = FALSE;
	}
#endif
#if XM7_VER == 1 && defined(SFDC)
	if (!sfdc_load(fileh, ver)) {
		flag = FALSE;
	}
#endif	/* XM7_VER == 1 && defined(SFDC) */
#ifdef MOUSE
	if (!ptm_load(fileh, ver)) {
		flag = FALSE;
	}
#endif
	file_close(fileh);

#if XM7_VER >= 3
	/* 400C[ĥVRAMzu␳Kv */
	if (mode400l) {
		if (!fix_vram_address()) {
			flag = FALSE;
		}
	}
#endif

	/* ʍĕ` */
	display_notify();

	/* CPUx䗦ݒ */
	speed_ratio = 10000;

	if (!flag) {
		return STATELOAD_ERROR;
	}

	return STATELOAD_SUCCESS;
}

/*
 *	t@Cǂݍ(BYTE)
 */
BOOL FASTCALL file_byte_read(int fileh, BYTE *dat)
{
	return file_read(fileh, dat, 1);
}

/*
 *	t@Cǂݍ(WORD)
 */
BOOL FASTCALL file_word_read(int fileh, WORD *dat)
{
	BYTE tmp;

	if (!file_read(fileh, &tmp, 1)) {
		return FALSE;
	}
	*dat = tmp;

	if (!file_read(fileh, &tmp, 1)) {
		return FALSE;
	}
	*dat <<= 8;
	*dat |= tmp;

	return TRUE;
}

/*
 *	t@Cǂݍ(DWORD)
 */
BOOL FASTCALL file_dword_read(int fileh, DWORD *dat)
{
	BYTE tmp;

	if (!file_read(fileh, &tmp, 1)) {
		return FALSE;
	}
	*dat = tmp;

	if (!file_read(fileh, &tmp, 1)) {
		return FALSE;
	}
	*dat *= 256;
	*dat |= tmp;

	if (!file_read(fileh, &tmp, 1)) {
		return FALSE;
	}
	*dat *= 256;
	*dat |= tmp;

	if (!file_read(fileh, &tmp, 1)) {
		return FALSE;
	}
	*dat *= 256;
	*dat |= tmp;

	return TRUE;
}

/*
 *	t@Cǂݍ(BOOL)
 */
BOOL FASTCALL file_bool_read(int fileh, BOOL *dat)
{
	BYTE tmp;

	if (!file_read(fileh, &tmp, 1)) {
		return FALSE;
	}

	switch (tmp) {
		case 0:
			*dat = FALSE;
			return TRUE;
		case 0xff:
			*dat = TRUE;
			return TRUE;
	}

	return FALSE;
}

/*
 *	t@C(BYTE)
 */
BOOL FASTCALL file_byte_write(int fileh, BYTE dat)
{
	return file_write(fileh, &dat, 1);
}

/*
 *	t@C(WORD)
 */
BOOL FASTCALL file_word_write(int fileh, WORD dat)
{
	BYTE tmp;

	tmp = (BYTE)(dat >> 8);
	if (!file_write(fileh, &tmp, 1)) {
		return FALSE;
	}

	tmp = (BYTE)(dat & 0xff);
	if (!file_write(fileh, &tmp, 1)) {
		return FALSE;
	}

	return TRUE;
}

/*
 *	t@C(DWORD)
 */
BOOL FASTCALL file_dword_write(int fileh, DWORD dat)
{
	BYTE tmp;

	tmp = (BYTE)(dat >> 24);
	if (!file_write(fileh, &tmp, 1)) {
		return FALSE;
	}

	tmp = (BYTE)(dat >> 16);
	if (!file_write(fileh, &tmp, 1)) {
		return FALSE;
	}

	tmp = (BYTE)(dat >> 8);
	if (!file_write(fileh, &tmp, 1)) {
		return FALSE;
	}

	tmp = (BYTE)(dat & 0xff);
	if (!file_write(fileh, &tmp, 1)) {
		return FALSE;
	}

	return TRUE;
}

/*
 *	t@C(BOOL)
 */
BOOL FASTCALL file_bool_write(int fileh, BOOL dat)
{
	BYTE tmp;

	if (dat) {
		tmp = 0xff;
	}
	else {
		tmp = 0;
	}

	if (!file_write(fileh, &tmp, 1)) {
		return FALSE;
	}

	return TRUE;
}
